<?php

/**
 * InterWorx Hosting Control Panel
 *
 * <pre>
 * +----------------------------------------------------------------------+
 * | Copyright (c) 2000-2010 InterWorx L.L.C., All Rights Reserved.       |
 * +----------------------------------------------------------------------+
 * | Redistribution and use in source form, with or without modification  |
 * | is NOT permitted without consent from the copyright holder.          |
 * |                                                                      |
 * | THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS "AS IS" AND |
 * | ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,    |
 * | THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A          |
 * | PARTICULAR PURPOSE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    |
 * | EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,  |
 * | PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   |
 * | PROFITS; OF BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY  |
 * | OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT         |
 * | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE    |
 * | USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH     |
 * | DAMAGE.                                                              |
 * +----------------------------------------------------------------------+
 * </pre>
 *
 * @package InterWorx
 * @author  Malin Lorch <mlorch@liquidweb.com>
 * @date    Oct 21, 2020
 */

/**
 * Siteworx Daily Backups plugin.
 *
 * @package    InterWorx
 * @subpackage Plugin
 */
class Plugin_SiteworxBackupsCron extends Plugin
{
    const DEFAULT_BACKUPS_LOCATION = '/backup';
    const DEFAULT_KEEP_DAILY = 4;
    const DEFAULT_KEEP_WEEKLY = 3;
    const DEFAULT_KEEP_MONTHLY = 2;
    const DEFAULT_AVAIL = 5;
    const DEFAULT_BACKUP_SSHFS = '0';
    const DEFAULT_BACKUP_SSHFSUMOUNT = '0';


    const INPUT_DIRECTORY = 'directory';
    const INPUT_KEEP_DAILY = 'daily';
    const INPUT_KEEP_WEEKLY = 'weekly';
    const INPUT_KEEP_MONTHLY = 'monthly';
    const INPUT_SPACE = 'space';
    const INPUT_BACKUP_SSHFS = 'sshfs';
    const INPUT_BACKUP_SSHFSUMOUNT = 'sshfsumount';
    const INPUT_BACKUP_SSHFSUSER = 'sshfsuser';
    const INPUT_BACKUP_SSHFSPASS = 'sshfspass';
    const INPUT_BACKUP_SSHFSHOST = 'sshfshost';
    const INPUT_BACKUP_SSHFSPORT = 'sshfsport';
    const INPUT_BACKUP_SSHFSRDIR = 'rdir';

    private $_scriptPath;
    private $_ssfsconfPath;
    private $_crontab;
    private $_backupsCronJobs;
 
    /**
     * Decides which category to display plugin under.
     *
     * @return string
     */
    public function getCategory()
    {
        return Plugin_Category::BACKUP_RESTORE;
    }

    /**
     * Specify inputs etc to show when the plugin is edited.
     *
     * @param Form_NW_Plugins_Edit $Form
     */
    public function initializeMyEditForm(Form_NW_Plugins_Edit $Form)
    {
        $pluginsPath = Ini::get(Ini::IWORX_PLUGINS, 'dir');
        $this->_scriptPath = "$pluginsPath/siteworx-backups-cron/scripts/iwx-backup.sh";
        $this->_ssfsconfPath = "$pluginsPath/siteworx-backups-cron/scripts/iwx-sshfs.conf";
        $this->_crontab = IWorxCron::getUserCrontab('root', true);
        $this->_backupsCronJobs = $this->_getBackupsCronJobs();
        $this->_configureForm($Form);
    }


    /**
     * Check if fuse-sshfs is already installed and or available
     *
     * @return boolean
     */
    public function _fusesshfs()
    {
        $sshfsisinstalled = Yum::isInstalled("fuse-sshfs");
        if ($sshfsisinstalled == true){
            return $sshfsisinstalled;
        } else {
            $yum_info = Yum::runYumInfo("fuse-sshfs");
            return $yum_info[0]->name === 'fuse-sshfs';
        }
    }

    /**
     * Configure form to include input for backups path &
     * keep daily/weekly/monthly
     *
     * @param Form_NW_Plugins_Edit $Form
     */
    private function _configureForm(Form_NW_Plugins_Edit $Form)
    {

        $defaultValues = $this->_getDefaultValues();
        $defaultDir = $defaultValues[self::INPUT_DIRECTORY];
        $defaultKeepDaily = $defaultValues[self::INPUT_KEEP_DAILY];
        $defaultKeepWeekly = $defaultValues[self::INPUT_KEEP_WEEKLY];
        $defaultKeepMonthly = $defaultValues[self::INPUT_KEEP_MONTHLY];
        $defaultSpace = $defaultValues[self::INPUT_SPACE];
        $defaultsshfs = $defaultValues[self::INPUT_BACKUP_SSHFS];
        $defaultsshfsumount = $defaultValues[self::INPUT_BACKUP_SSHFSUMOUNT];
        $sshfsuser = $defaultValues[self::INPUT_BACKUP_SSHFSUSER];
        $sshfspass = $defaultValues[self::INPUT_BACKUP_SSHFSPASS];
        $sshfshost = $defaultValues[self::INPUT_BACKUP_SSHFSHOST];
        $sshfsport = $defaultValues[self::INPUT_BACKUP_SSHFSPORT];
        $sshfsrdir = $defaultValues[self::INPUT_BACKUP_SSHFSRDIR];

        $Group = $Form->addGroup('config', '##LG_CONFIGURATION##');

        $DirInput = new Input_String('directory', $defaultDir);
        $DirInput->addValidator(new IWorx_Validate_Directory_Path());

        $Group->addInput($DirInput)
        ->setLabel('Backups Path');

        $Group->addInput(new Input_Number(self::INPUT_KEEP_DAILY, $defaultKeepDaily))
        ->setLabel('Keep Daily Backups');

        $Group->addInput(new Input_Number(self::INPUT_KEEP_WEEKLY, $defaultKeepWeekly))
        ->setLabel('Keep Weekly Backups');

        $Group->addInput(new Input_Number(self::INPUT_KEEP_MONTHLY, $defaultKeepMonthly))
        ->setLabel('Keep Monthly Backups');

        $Group->addInput(new Input_Number(self::INPUT_SPACE, $defaultSpace))
        ->setLabel('Minimum Available Disk Space (%)');

        $dstype = DataSource_Boolean::YES_NO;
        $exsshfs    = ( $defaultsshfs == 1 ) ? 1 : 0;
        $yum_info = $this->_fusesshfs();
        if ($yum_info === true) {
            $Showsshfs = new Input_Select(self::INPUT_BACKUP_SSHFS, $exsshfs);
            $Showsshfs->setDataSource(DataSource_Boolean::factory($dstype))
            ->setLabel('Enable Remote SSHFS Mount');

            $Group->addInput($Showsshfs);

            $Group->addInput(new Input_String(self::INPUT_BACKUP_SSHFSUSER, $sshfsuser))
            ->setLabel('SSH or SCP User')
            ->setShowOn($Showsshfs, '1');

            $Group->addInput(new Input_Password(self::INPUT_BACKUP_SSHFSPASS, $sshfspass))
            ->setLabel('SSH or SCP User Password')
            ->setShowOn($Showsshfs, '1');

            $Group->addInput(new Input_HostOrIP(self::INPUT_BACKUP_SSHFSHOST, $sshfshost))
            ->setLabel('SSH or SCP Host or IP')
            ->setShowOn($Showsshfs, '1');

            $Group->addInput(new Input_Number(self::INPUT_BACKUP_SSHFSPORT, $sshfsport))
            ->setLabel('SSH or SCP Port')
            ->setShowOn($Showsshfs, '1');

            $Group->addInput(new Input_String(self::INPUT_BACKUP_SSHFSRDIR, $sshfsrdir))
            ->setLabel('SSH remote backup folder')
            ->setShowOn($Showsshfs, '1');

            $Group->addInput(new Input_Boolean_YesNo(self::INPUT_BACKUP_SSHFSUMOUNT, $defaultsshfsumount))
            ->setLabel('Unmount remote sshfs volume when backup is complete')
            ->setShowOn($Showsshfs, '1');

        } else {
            $Group->addInput(new Input_String("SSHFS", "Remote SSH or SCP backups not possible"))
            ->disable()
            ->setLabel('SSHFS Available?');
        }
    }

    /**
     * Get default values by parsing script command.
     *
     * @return array
     */
    private function _getDefaultValues()
    {
        $keepDaily = self::DEFAULT_KEEP_DAILY;
        $keepWeekly = self::DEFAULT_KEEP_WEEKLY;
        $keepMonthly = self::DEFAULT_KEEP_MONTHLY;
        $path = self::DEFAULT_BACKUPS_LOCATION;
        $space = self::DEFAULT_AVAIL;
        $sshfs = self::DEFAULT_BACKUP_SSHFS;
        $sshfsumount = self::DEFAULT_BACKUP_SSHFSUMOUNT;

        $sshfs_array = parse_ini_file("/usr/local/interworx/plugins/siteworx-backups-cron/scripts/iwx-sshfs.conf");
        $sshfsuser = $sshfs_array["user"];
        $sshfspass = $sshfs_array["pass"];
        $sshfshost = $sshfs_array["host"];
        $sshfsport = $sshfs_array["port"];
        $sshfsrdir = $sshfs_array["rdir"];

        if ($this->_cronJobExists()) {
            $defaults = explode(' ', $this->_backupsCronJobs[0]['script']);
            $keepDaily = $defaults[2];
            $keepWeekly = $defaults[3];
            $keepMonthly = $defaults[4];
            $path = $defaults[5];
            $space = $defaults[6];
            $sshfs = $defaults[7];
            $sshfsumount = $defaults[8];
        }

        return array(
        self::INPUT_DIRECTORY => $path,
        self::INPUT_SPACE => $space,
        self::INPUT_KEEP_DAILY => $keepDaily,
        self::INPUT_KEEP_WEEKLY => $keepWeekly,
        self::INPUT_KEEP_MONTHLY => $keepMonthly,
        self::INPUT_BACKUP_SSHFS => $sshfs,
        self::INPUT_BACKUP_SSHFSUMOUNT => $sshfsumount,
        self::INPUT_BACKUP_SSHFSUSER => $sshfsuser,
        self::INPUT_BACKUP_SSHFSPASS => $sshfspass,
        self::INPUT_BACKUP_SSHFSPORT => $sshfsport,
        self::INPUT_BACKUP_SSHFSHOST => $sshfshost,
        self::INPUT_BACKUP_SSHFSRDIR => $sshfsrdir
        );
    }

    /**
     * Get cron jobs executing backup script.
     *
     * @return array
     */
    private function _getBackupsCronJobs()
    {
        if (isset($this->_crontab) && !empty($this->_crontab['jobs'])) {
            return $this->_getDuplicateJobs($this->_crontab['jobs']);
        }
    }

    /**
     * @param array $cronJobs
     * @return array
     */
    private function _getDuplicateJobs(array $cronJobs)
    {
        $duplicates = array();

        foreach ($cronJobs as $cronJob) {
            if (strpos($cronJob['script'], $this->_scriptPath) !== false) {
                $duplicates[] = $cronJob;
            }
        }

        return $duplicates;
    }

    /**
     * Add task to daily cron if plugin is being enabled.
     *
     * @param Form $Form
     */
    public function validateMyEditForm(Form $Form)
    {
        try {
            $this->_processRequest($Form);
            return true;
        } catch (Exception $e) {
            return false;
        }
    }

    /**
     * Process request as part of validate
     * in order to be able to catch any errors.
     *
     * @param Form_NW_Plugins_Edit $Form
     */
    private function _processRequest(Form_NW_Plugins_Edit $Form)
    {
        if ($Form->enableRequest()) {
            $this->_setUpCronJob($Form);
        }

        if (!$Form->enableRequest() && $this->_cronJobExists()) {
            $this->_deleteBackupsCronJobs();
        }
    }

    /**
     * Set up cron job.
     *
     * @param Form_NW_Plugins_Edit $Form
     */
    private function _setUpCronJob(Form_NW_Plugins_Edit $Form)
    {
        if ($this->_cronJobExists()) {
            $this->_deleteBackupsCronJobs();
        }

        $this->_addCronJob($Form);
    }

    /**
     * @return boolean
     */
    private function _cronJobExists()
    {
        return !empty($this->_backupsCronJobs);
    }

    /**
     * add cron job.
     *
     * @param Form_NW_Plugins_Edit $Form
     */
    private function _addCronJob(Form_NW_Plugins_Edit $Form)
    {
        $backupsPath = $Form->getValue(self::INPUT_DIRECTORY);
        $usageSpace = $Form->getValue(self::INPUT_SPACE);
        $keepDaily = $Form->getValue(self::INPUT_KEEP_DAILY);
        $keepWeekly = $Form->getValue(self::INPUT_KEEP_WEEKLY);
        $keepMonthly = $Form->getValue(self::INPUT_KEEP_MONTHLY);
        $sshfs = $Form->getValue(self::INPUT_BACKUP_SSHFS);
        $sshfsumount = $Form->getValue(self::INPUT_BACKUP_SSHFSUMOUNT);

        #Update sshfs confs.
        $sshfs_array = parse_ini_file("/usr/local/interworx/plugins/siteworx-backups-cron/scripts/iwx-sshfs.conf");
        $sshfsuser = $sshfs_array["user"];
        $sshfsusernew = $Form->getValue(self::INPUT_BACKUP_SSHFSUSER);
        $sshfspass = $sshfs_array["pass"];
        $sshfspassnew = $Form->getValue(self::INPUT_BACKUP_SSHFSPASS);
        $sshfshost = $sshfs_array["host"];
        $sshfshostnew = $Form->getValue(self::INPUT_BACKUP_SSHFSHOST);
        $sshfsport = $sshfs_array["port"];
        $sshfsportnew = $Form->getValue(self::INPUT_BACKUP_SSHFSPORT);
        $sshfsrdir = $sshfs_array["rdir"];
        $sshfsrdirnew = $Form->getValue(self::INPUT_BACKUP_SSHFSRDIR);
        $sshfsconfin = file_get_contents($this->_ssfsconfPath);
        $sshfsconfout = preg_replace("'".$sshfsuser."'", $sshfsusernew, $sshfsconfin);
        $sshfsconfout = preg_replace("'".$sshfspass."'", $sshfspassnew, $sshfsconfout);
        $sshfsconfout = preg_replace("'".$sshfsport."'", $sshfsportnew, $sshfsconfout);
        $sshfsconfout = preg_replace("'".$sshfshost."'", $sshfshostnew, $sshfsconfout);
        $sshfsconfout = preg_replace("'".$sshfsrdir."'", $sshfsrdirnew, $sshfsconfout);
        file_put_contents($this->_ssfsconfPath, $sshfsconfout);


        $args = array(
        'user' => 'root',
        'hour' => array('3'),
        'minute' => array('0'),
        'day' => '*', // every day
        'month' => '*', // every month
        'script' => "sh $this->_scriptPath $keepDaily $keepWeekly $keepMonthly $backupsPath $usageSpace $sshfs $sshfsumount",
        );

        #Install fuse-sshfs if sshfs enabled and fuse-sshfs not already installed.
        if ($sshfs == 1) {
            $fusesshfsinst = Yum::isInstalled("fuse-sshfs");
            if ($fusesshfsinst === false){
                Yum::install("fuse-sshfs");
            }
        }

        $Route = IW::FC()->routeFromPHP(
            'Ctrl_Nodeworx_Cron',
            'addCommit',
            $args,
            IW::NW()
        );

        $errors = $Route->getErrors();

        if (!empty($errors)) {
            $message = 'Failed to set up cron job';
            IW::getLog()->error($message, $errors);

            throw new IWorx_Exception_Plugin('Failed to set up cron job');
        }
    }


    /**
     * Delete cron jobs running backups script.
     */
    private function _deleteBackupsCronJobs()
    {
        foreach ($this->_backupsCronJobs as $cronJob) {
            $linenum = $cronJob['linenum'];
            unset($this->_crontab['jobs'][$linenum]);
        }

        $retval = IWorxCron::installCrontab($this->_crontab, 'root', $result);

        if ($retval != CLI::OK) {
            $message = 'Failed deleting cron job';
            IW::getLog()->error($message);
            throw new IWorx_Exception_Plugin($message);
        }
    }
}
